from .base import BootstrapTestCase


class BootstrapAlertTestCase(BootstrapTestCase):
    def test_bootstrap_alert(self):
        self.assertEqual(
            self.render('{% bootstrap_alert "content" dismissible=False %}', None),
            '<div class="alert alert-info" role="alert">content</div>',
        )

    def test_bootstrap_alert_type_invalid(self):
        with self.assertRaises(ValueError):
            self.render('{% bootstrap_alert "content" alert_type="nope" %}', None)

    def test_bootstrap_alert_dismissible(self):
        self.assertEqual(
            self.render('{% bootstrap_alert "content" %}'),
            (
                '<div class="alert alert-info alert-dismissible fade show" role="alert">'
                "content"
                '<button type="button" class="btn-close" data-bs-dismiss="alert" aria-label="Close"></button>'
                "</div>"
            ),
        )

    def test_bootstrap_alert_html_content(self):
        self.assertEqual(
            self.render('{% bootstrap_alert "foo<br>bar" dismissible=False %}'),
            '<div class="alert alert-info" role="alert">foo<br>bar</div>',
        )
        self.assertEqual(
            self.render("{% bootstrap_alert value dismissible=False %}", context={"value": "foo<br>bar"}),
            '<div class="alert alert-info" role="alert">foo&lt;br&gt;bar</div>',
        )
        self.assertEqual(
            self.render("{% bootstrap_alert value|safe dismissible=False %}", context={"value": "foo<br>bar"}),
            '<div class="alert alert-info" role="alert">foo<br>bar</div>',
        )



from django import forms

from .base import BootstrapTestCase


class SelectTestForm(forms.Form):
    test = forms.ChoiceField(
        choices=(
            (1, "one"),
            (2, "two"),
        )
    )


class BootstrapFieldSelectTestCase(BootstrapTestCase):
    def test_select(self):
        """Test field with select widget."""
        self.assertHTMLEqual(
            self.render("{% bootstrap_field form.test %}", context={"form": SelectTestForm()}),
            (
                '<div class="django_bootstrap5-req mb-3">'
                '<label for="id_test" class="form-label">Test</label>'
                '<select class="form-select" id="id_test" name="test">'
                '<option value="1">one</option>'
                '<option value="2">two</option>'
                "</select>"
                "</div>"
            ),
        )

    def test_select_horizontal(self):
        """Test field with select widget in horizontal layout."""
        self.assertHTMLEqual(
            self.render('{% bootstrap_field form.test layout="horizontal" %}', context={"form": SelectTestForm()}),
            (
                '<div class="django_bootstrap5-req row mb-3">'
                '<label for="id_test" class="col-form-label col-sm-2">Test</label>'
                '<div class="col-sm-10">'
                '<select class="form-select" id="id_test" name="test">'
                '<option value="1">one</option>'
                '<option value="2">two</option>'
                "</select>"
                "</div>"
                "</div>"
            ),
        )

    def test_select_floating(self):
        """Test field with select widget in floating layout."""
        self.assertHTMLEqual(
            self.render('{% bootstrap_field form.test layout="floating" %}', context={"form": SelectTestForm()}),
            (
                '<div class="django_bootstrap5-req form-floating mb-3">'
                '<select class="form-select" id="id_test" name="test">'
                '<option value="1">one</option>'
                '<option value="2">two</option>'
                "</select>"
                '<label for="id_test" class="form-label">Test</label>'
                "</div>"
            ),
        )



from django.test import TestCase

from django_bootstrap5.css import merge_css_classes


class CssTestCase(TestCase):
    def test_merge_css_classes(self):
        css_classes = "one two"
        css_class = "three four"

        classes = merge_css_classes(css_classes, css_class)
        self.assertEqual(classes, "one two three four")

        classes = merge_css_classes(css_class, css_classes)
        self.assertEqual(classes, "three four one two")



from django import forms
from django.forms import TextInput

from tests.base import BootstrapTestCase


class InputTypeTextTestCase(BootstrapTestCase):
    """Test for TextInput widgets that only differ in type, and have no other exceptions."""

    def _html_default(self, input_type):
        """Output for input with given input_type in default layout."""
        return (
            '<div class="django_bootstrap5-req mb-3">'
            '<label for="id_test" class="form-label">Test</label>'
            f'<input class="form-control" id="id_test" name="test" placeholder="Test" required type="{input_type}">'
            "</div>"
        )

    def _html_horizontal(self, input_type):
        """Output for input with given input_type in horizontal layout."""
        return (
            '<div class="django_bootstrap5-req mb-3 row">'
            '<label class="col-form-label col-sm-2" for="id_test">'
            "Test"
            '</label><div class="col-sm-10">'
            f'<input class="form-control" id="id_test" name="test" placeholder="Test" required type="{input_type}">'
            "</div>"
            "</div>"
        )

    def _html_floating(self, input_type):
        """
        Output for input with given input_type in floating layout.

        Note that this function does not check if the flaoting layout is supported for this input_type.
        """
        return (
            '<div class="django_bootstrap5-req mb-3 form-floating">'
            f'<input class="form-control" id="id_test" name="test" placeholder="Test" required type="{input_type}">'
            '<label for="id_test" class="form-label">Test</label>'
            "</div>"
        )

    def test_input_type_text(self):
        """Test field with default CharField widget."""

        class TextTestForm(forms.Form):
            test = forms.CharField()

        form = TextTestForm()

        self.assertHTMLEqual(
            self.render("{% bootstrap_field form.test %}", context={"form": form}),
            self._html_default("text"),
        )

        addon_html = (
            '<div class="django_bootstrap5-req mb-3">'
            '<label for="id_test" class="form-label">Test</label>'
            '<div class="input-group mb-3">'
            '<span class="input-group-text">foo</span>'
            '<input class="form-control" id="id_test" name="test" placeholder="Test" required type="text">'
            "</div>"
            "</div>"
        )

        self.assertHTMLEqual(
            self.render('{% bootstrap_field form.test addon_before="foo" %}', context={"form": form}),
            addon_html,
        )

        self.assertHTMLEqual(
            self.render('{% bootstrap_field form.test addon_before="foo" layout="floating" %}', context={"form": form}),
            addon_html,
        )

        self.assertHTMLEqual(
            self.render("{% bootstrap_field form.test layout='horizontal' %}", context={"form": form}),
            self._html_horizontal("text"),
        ),

        self.assertHTMLEqual(
            self.render("{% bootstrap_field form.test layout='floating' %}", context={"form": form}),
            self._html_floating("text"),
        )

    def _test_input_type(self, input_type):
        """Test field with given input type in all layouts."""

        class InputTypeTestForm(forms.Form):
            test = forms.CharField(widget=TextInput(attrs={"type": input_type}))

        form = InputTypeTestForm()
        default_html = self._html_default(input_type)
        horizontal_html = self._html_horizontal(input_type)
        floating_html = self._html_floating(input_type)

        self.assertHTMLEqual(
            self.render("{% bootstrap_field form.test %}", context={"form": form}),
            default_html,
        )

        self.assertHTMLEqual(
            self.render("{% bootstrap_field form.test layout='horizontal' %}", context={"form": form}),
            horizontal_html,
        )

        self.assertHTMLEqual(
            self.render("{% bootstrap_field form.test layout='floating' %}", context={"form": form}),
            floating_html,
        )

    def test_input_types(self):
        """Test field with CharField widget and its type set."""
        self._test_input_type("text")
        self._test_input_type("number")
        self._test_input_type("email")
        self._test_input_type("url")
        self._test_input_type("tel")
        self._test_input_type("date")
        self._test_input_type("time")
        self._test_input_type("password")

    def test_input_type_password(self):
        """Test field with password widget."""

        class PasswordTestForm(forms.Form):
            test = forms.CharField(widget=forms.PasswordInput)

        form = PasswordTestForm()

        self.assertHTMLEqual(
            self.render("{% bootstrap_field form.test %}", context={"form": form}),
            self._html_default("password"),
        )

        self.assertHTMLEqual(
            self.render("{% bootstrap_field form.test layout='horizontal' %}", context={"form": form}),
            self._html_horizontal("password"),
        ),

        self.assertHTMLEqual(
            self.render("{% bootstrap_field form.test layout='floating' %}", context={"form": form}),
            self._html_floating("password"),
        )



from bs4 import BeautifulSoup
from django import forms
from django.forms import formset_factory

from django_bootstrap5.exceptions import BootstrapError
from tests.base import BootstrapTestCase


class FormTestForm(forms.Form):
    required_text = forms.CharField(required=True, help_text="<i>required_text_help</i>")
    optional_text = forms.CharField(required=False, help_text="<i>required_text_help</i>")


class ShowLabelTestForm(forms.Form):
    subject = forms.CharField()


class NonFieldErrorTestForm(FormTestForm):
    non_field_error_message = "This is a non field error."

    def clean(self):
        super().clean()
        raise forms.ValidationError(self.non_field_error_message)


class BootstrapFormTestCase(BootstrapTestCase):
    def test_illegal_form(self):
        with self.assertRaises(BootstrapError):
            self.render("{% bootstrap_form form %}", {"form": "illegal"})

    def test_exclude(self):
        html = self.render('{% bootstrap_form form exclude="optional_text" %}', {"form": FormTestForm()})
        self.assertNotIn("optional_text", html)

    def test_error_class(self):
        form = FormTestForm({"optional_text": "my_message"})

        html = self.render("{% bootstrap_form form %}", {"form": form})
        self.assertIn("django_bootstrap5-err", html)

        html = self.render('{% bootstrap_form form error_css_class="custom-error-class" %}', {"form": form})
        self.assertIn("custom-error-class", html)

        html = self.render('{% bootstrap_form form error_css_class="" %}', {"form": form})
        self.assertNotIn("django_bootstrap5-err", html)

    def test_required_class(self):
        form = FormTestForm({"subject": "subject"})
        html = self.render("{% bootstrap_form form %}", {"form": form})
        self.assertIn("django_bootstrap5-req", html)

        html = self.render('{% bootstrap_form form required_css_class="custom-required-class" %}', {"form": form})
        self.assertIn("custom-required-class", html)

        html = self.render('{% bootstrap_form form required_css_class="" %}', {"form": form})
        self.assertNotIn("django_bootstrap5-req", html)

    def test_bound_class(self):
        form = FormTestForm({"subject": "subject"})

        html = self.render("{% bootstrap_form form %}", {"form": form})
        self.assertIn("django_bootstrap5-bound", html)

        form = FormTestForm({"subject": "subject"})

        html = self.render('{% bootstrap_form form bound_css_class="successful-test" %}', {"form": form})
        self.assertIn("successful-test", html)

        form = FormTestForm({"subject": "subject"})

        html = self.render('{% bootstrap_form form bound_css_class="" %}', {"form": form})
        self.assertNotIn("django_bootstrap5-bound", html)

    def test_alert_error_type(self):
        form = NonFieldErrorTestForm({"subject": "subject"})

        html = self.render("{% bootstrap_form form alert_error_type='all' %}", {"form": form})
        soup = BeautifulSoup(html, "html.parser")
        errors = list(soup.select(".text-danger")[0].stripped_strings)
        self.assertIn(form.non_field_error_message, errors)
        self.assertIn("This field is required.", errors)

        html = self.render("{% bootstrap_form form alert_error_type='non_fields' %}", {"form": form})
        self.assertEqual(
            html,
            self.render("{% bootstrap_form form %}", {"form": form}),
            "Default behavior is not the same as showing non-field errors",
        )

        soup = BeautifulSoup(html, "html.parser")
        errors = list(soup.select(".text-danger")[0].stripped_strings)
        self.assertIn(form.non_field_error_message, errors)
        self.assertNotIn("This field is required.", errors)

        html = self.render("{% bootstrap_form form alert_error_type='fields' %}", {"form": form})
        soup = BeautifulSoup(html, "html.parser")
        errors = list(soup.select(".text-danger")[0].stripped_strings)
        self.assertNotIn(form.non_field_error_message, errors)
        self.assertIn("This field is required.", errors)

        html = self.render("{% bootstrap_form form alert_error_type='none' %}", {"form": form})
        soup = BeautifulSoup(html, "html.parser")
        self.assertFalse(soup.select(".text-danger"))

    def test_form_errors(self):
        form = FormTestForm({"subject": "subject"})
        html = self.render("{% bootstrap_form_errors form %}", {"form": form})
        self.assertHTMLEqual(
            html,
            '<ul class="list-unstyled text-danger"><li>This field is required.</li></ul>',
        )


class ShowLabelTestCase(BootstrapTestCase):
    def test_show_label_false(self):
        self.assertInHTML(
            '<label class="visually-hidden" for="id_subject">Subject</label>',
            self.render("{% bootstrap_form form show_label=False %}", {"form": ShowLabelTestForm()}),
        )

    def test_show_label_sr_only(self):
        self.assertInHTML(
            '<label class="visually-hidden" for="id_subject">Subject</label>',
            self.render("{% bootstrap_form form show_label='' %}", {"form": ShowLabelTestForm()}),
        )

    def test_show_label_skip(self):
        self.assertNotIn(
            "label",
            self.render("{% bootstrap_form form show_label='skip' %}", {"form": ShowLabelTestForm()}),
        )

    def test_show_label_false_in_formset(self):
        TestFormSet = formset_factory(ShowLabelTestForm, extra=1)
        self.assertInHTML(
            '<label class="visually-hidden" for="id_form-0-subject">Subject</label>',
            self.render("{% bootstrap_formset formset show_label=False %}", {"formset": TestFormSet()}),
        )



from django_bootstrap5.exceptions import BootstrapError
from tests.base import BootstrapTestCase


class ButtonTestCase(BootstrapTestCase):
    def test_button(self):
        self.assertHTMLEqual(
            self.render("{% bootstrap_button 'button' size='lg' %}"),
            '<button class="btn btn-primary btn-lg">button</button>',
        )

        self.assertHTMLEqual(
            self.render("{% bootstrap_button 'button' id='foo' %}"),
            '<button class="btn btn-primary" id="foo">button</button>',
        )

        self.assertHTMLEqual(
            self.render("{% bootstrap_button 'button' name='foo' %}"),
            '<button class="btn btn-primary" name="foo">button</button>',
        )

        self.assertHTMLEqual(
            self.render("{% bootstrap_button 'button' value='foo' %}"),
            '<button class="btn btn-primary" value="foo">button</button>',
        )

        self.assertHTMLEqual(
            self.render("{% bootstrap_button 'button' title='foo' %}"),
            '<button class="btn btn-primary" title="foo">button</button>',
        )

    def test_button_type_link(self):
        link_button = '<a class="btn btn-primary btn-lg" href="#" role="button">button</a>'
        self.assertHTMLEqual(
            self.render("{% bootstrap_button 'button' size='lg' href='#' %}"),
            link_button,
        )
        self.assertHTMLEqual(
            self.render("{% bootstrap_button 'button' button_type='link' size='lg' href='#' %}"),
            link_button,
        )

        with self.assertRaises(BootstrapError):
            self.render("{% bootstrap_button 'button' button_type='button' href='#' %}")



from tests.base import BootstrapTestCase


class TemplateTestCase(BootstrapTestCase):
    def test_empty_template(self):
        self.assertEqual(self.render("").strip(), "")

    def test_text_template(self):
        self.assertEqual(self.render("some text").strip(), "some text")

    def test_bootstrap5_html_template_title(self):
        html = self.render(
            '{% extends "django_bootstrap5/bootstrap5.html" %}{% block bootstrap5_title %}x-title-x{% endblock %}',
            load_bootstrap=False,
        )
        self.assertIn("x-title-x", html)

    def test_bootstrap5_html_template_content(self):
        html = self.render(
            '{% extends "django_bootstrap5/bootstrap5.html" %}{% block bootstrap5_content %}x-content-x{% endblock %}',
            load_bootstrap=False,
        )
        self.assertIn("x-content-x", html)



def _css_class_list(list_of_css_classes):
    """Return list without duplicate or empty elements."""
    return filter(None, list(dict.fromkeys(list_of_css_classes)))


def _css_class_list_string(list_of_css_classes):
    """Return string version of list without duplicate or empty elements."""
    return " ".join(_css_class_list(list_of_css_classes))


def merge_css_classes(*args):
    """Merge CSS classes into one string."""
    css_classes = []
    for arg in args:
        css_classes += f"{arg}".split(" ")
    return _css_class_list_string(css_classes)



from django.test import TestCase

from django_bootstrap5.core import get_bootstrap_setting


class SettingsTestCase(TestCase):
    def test_get_bootstrap_setting(self):
        self.assertIsNone(get_bootstrap_setting("SETTING_DOES_NOT_EXIST"))
        self.assertEqual("not none", get_bootstrap_setting("SETTING_DOES_NOT_EXIST", "not none"))
        with self.settings(BOOTSTRAP5={"SETTING_DOES_NOT_EXIST": "exists now"}):
            self.assertEqual(get_bootstrap_setting("SETTING_DOES_NOT_EXIST"), "exists now")



from django.utils.safestring import mark_safe

from .core import get_field_renderer, get_form_renderer, get_formset_renderer
from .css import merge_css_classes
from .exceptions import BootstrapError
from .html import render_tag
from .size import DEFAULT_SIZE, SIZE_MD, get_size_class

WRAPPER_CLASS = ""
WRAPPER_TAG = "div"


def render_formset(formset, **kwargs):
    """Render a formset to a Bootstrap layout."""
    renderer_cls = get_formset_renderer(**kwargs)
    return renderer_cls(formset, **kwargs).render()


def render_formset_errors(formset, **kwargs):
    """Render formset errors to a Bootstrap layout."""
    renderer_cls = get_formset_renderer(**kwargs)
    return renderer_cls(formset, **kwargs).render_errors()


def render_form(form, **kwargs):
    """Render a form to a Bootstrap layout."""
    renderer_cls = get_form_renderer(**kwargs)
    return renderer_cls(form, **kwargs).render()


def render_form_errors(form, type="all", **kwargs):
    """Render form errors to a Bootstrap layout."""
    renderer_cls = get_form_renderer(**kwargs)
    return renderer_cls(form, **kwargs).render_errors(type)


def render_field(field, **kwargs):
    """Render a field to a Bootstrap layout."""
    renderer_cls = get_field_renderer(**kwargs)
    return renderer_cls(field, **kwargs).render()


def render_label(content, label_for=None, label_class=None, label_title=""):
    """Render a label with content."""
    attrs = {}
    if label_for:
        attrs["for"] = label_for
    if label_class:
        attrs["class"] = label_class
    if label_title:
        attrs["title"] = label_title
    return render_tag("label", attrs=attrs, content=content)


def render_button(
    content,
    button_type=None,
    button_class="btn-primary",
    button_outline=False,
    size="",
    href="",
    name=None,
    value=None,
    title=None,
    extra_classes="",
    id="",
):
    """Render a button with content."""
    attrs = {}
    size_class = get_size_class(size, prefix="btn", skip=SIZE_MD, default=DEFAULT_SIZE)
    classes = merge_css_classes("btn", button_class, size_class)

    if button_type:
        if button_type not in ("submit", "reset", "button", "link"):
            raise BootstrapError(
                (
                    'Parameter "button_type" should be "submit", "reset", "button", "link" or empty '
                    '("{button_type}" given).'
                ).format(button_type=button_type)
            )
        if button_type != "link":
            attrs["type"] = button_type

    classes = merge_css_classes(classes, extra_classes)
    attrs["class"] = classes

    if href:
        tag = "a"
        if button_type and button_type != "link":
            raise BootstrapError(
                'Button of type "{button_type}" is not allowed a "href" parameter.'.format(button_type=button_type)
            )
        attrs["href"] = href
        # Specify role for link with button appearance
        attrs.setdefault("role", "button")
    else:
        tag = "button"

    if id:
        attrs["id"] = id
    if name:
        attrs["name"] = name
    if value:
        attrs["value"] = value
    if title:
        attrs["title"] = title
    return render_tag(tag, attrs=attrs, content=mark_safe(content))



from django import forms

from .base import BootstrapTestCase


class TestForm(forms.Form):
    test = forms.CharField(widget=forms.Textarea)


class BootstrapFieldTestCase(BootstrapTestCase):
    def test_textarea(self):
        """Test field with textarea widget."""
        self.assertHTMLEqual(
            self.render("{% bootstrap_field form.test %}", context={"form": TestForm()}),
            (
                '<div class="django_bootstrap5-req mb-3">'
                '<label for="id_test" class="form-label">Test</label>'
                '<textarea class="form-control" cols="40" id="id_test" name="test" placeholder="Test"'
                ' required rows="10"></textarea>'
                "</div>"
            ),
        )

    def test_textarea_floating(self):
        """Test field with textarea widget and floating label."""
        self.assertHTMLEqual(
            self.render('{% bootstrap_field form.test layout="floating" %}', context={"form": TestForm()}),
            (
                '<div class="django_bootstrap5-req form-floating mb-3">'
                '<textarea class="form-control" cols="40" id="id_test" name="test" placeholder="Test"'
                ' required rows="10"></textarea>'
                '<label for="id_test" class="form-label">Test</label>'
                "</div>"
            ),
        )



from django.forms import (
    BaseForm,
    BaseFormSet,
    BoundField,
    CheckboxInput,
    CheckboxSelectMultiple,
    ClearableFileInput,
    MultiWidget,
    RadioSelect,
    Select,
)
from django.forms.widgets import Input, SelectMultiple, Textarea
from django.utils.html import conditional_escape, format_html, strip_tags
from django.utils.safestring import mark_safe

from .core import get_bootstrap_setting
from .css import merge_css_classes
from .exceptions import BootstrapError
from .forms import WRAPPER_CLASS, WRAPPER_TAG, render_field, render_form, render_label
from .size import DEFAULT_SIZE, SIZE_MD, SIZE_XS, get_size_class, parse_size
from .text import text_value
from .utils import render_template_file
from .widgets import ReadOnlyPasswordHashWidget, is_widget_with_placeholder


class BaseRenderer(object):
    """A content renderer."""

    def __init__(self, *args, **kwargs):
        self.layout = kwargs.get("layout", "")
        self.wrapper_class = kwargs.get("wrapper_class", WRAPPER_CLASS)
        self.field_class = kwargs.get("field_class", "")
        self.label_class = kwargs.get("label_class", "")
        self.show_help = kwargs.get("show_help", True)
        self.show_label = kwargs.get("show_label", True)
        self.exclude = kwargs.get("exclude", "")
        self.set_placeholder = kwargs.get("set_placeholder", True)
        self.size = parse_size(kwargs.get("size", ""), default=SIZE_MD)
        self.horizontal_label_class = kwargs.get(
            "horizontal_label_class", get_bootstrap_setting("horizontal_label_class")
        )
        self.horizontal_field_class = kwargs.get(
            "horizontal_field_class", get_bootstrap_setting("horizontal_field_class")
        )
        self.checkbox_layout = kwargs.get("checkbox_layout", get_bootstrap_setting("checkbox_layout"))
        self.checkbox_style = kwargs.get("checkbox_style", get_bootstrap_setting("checkbox_style"))
        self.horizontal_field_offset_class = kwargs.get(
            "horizontal_field_offset_class", get_bootstrap_setting("horizontal_field_offset_class")
        )
        self.inline_field_class = kwargs.get("inline_field_class", get_bootstrap_setting("inline_field_class"))
        self.error_css_class = kwargs.get("error_css_class", None)
        self.required_css_class = kwargs.get("required_css_class", None)
        self.bound_css_class = kwargs.get("bound_css_class", None)
        self.alert_error_type = kwargs.get("alert_error_type", "non_fields")

    @property
    def is_floating(self):
        """Return whether to render `form-control` widgets as floating."""
        return self.layout == "floating"

    @property
    def is_horizontal(self):
        """Return whether to render form horizontally."""
        return self.layout == "horizontal"

    @property
    def is_inline(self):
        """Return whether to render widgets with inline layout."""
        return self.layout == "inline"

    def parse_size(self, size):
        """Return size if it is valid, default size if size is empty, or throws exception."""
        size = parse_size(size, default=DEFAULT_SIZE)
        if size == SIZE_XS:
            raise BootstrapError('Size "xs" is not valid for form controls.')
        return size

    def get_size_class(self, prefix):
        """Return size class for given prefix."""
        return get_size_class(self.size, prefix=prefix) if self.size in ["sm", "lg"] else ""

    def get_kwargs(self):
        """Return kwargs to pass on to child renderers."""
        context = {
            "layout": self.layout,
            "wrapper_class": self.wrapper_class,
            "field_class": self.field_class,
            "label_class": self.label_class,
            "show_help": self.show_help,
            "show_label": self.show_label,
            "exclude": self.exclude,
            "set_placeholder": self.set_placeholder,
            "size": self.size,
            "horizontal_label_class": self.horizontal_label_class,
            "horizontal_field_class": self.horizontal_field_class,
            "checkbox_layout": self.checkbox_layout,
            "checkbox_style": self.checkbox_style,
            "inline_field_class": self.inline_field_class,
            "error_css_class": self.error_css_class,
            "bound_css_class": self.bound_css_class,
            "required_css_class": self.required_css_class,
            "alert_error_type": self.alert_error_type,
        }
        return context

    def get_context_data(self):
        """Return context data for rendering."""
        return self.get_kwargs()

    def render(self):
        """Render to string."""
        return ""


class FormsetRenderer(BaseRenderer):
    """Default formset renderer."""

    def __init__(self, formset, *args, **kwargs):
        if not isinstance(formset, BaseFormSet):
            raise BootstrapError('Parameter "formset" should contain a valid Django Formset.')
        self.formset = formset
        super().__init__(*args, **kwargs)

    def get_context_data(self):
        context = super().get_context_data()
        context["formset"] = self.formset
        return context

    def render_management_form(self):
        """Return HTML for management form."""
        return text_value(self.formset.management_form)

    def render_forms(self):
        rendered_forms = mark_safe("")
        kwargs = self.get_kwargs()
        for form in self.formset.forms:
            rendered_forms += render_form(form, **kwargs)
        return rendered_forms

    def get_formset_errors(self):
        return self.formset.non_form_errors()

    def render_errors(self):
        formset_errors = self.get_formset_errors()
        if formset_errors:
            return render_template_file(
                "django_bootstrap5/form_errors.html",
                context={
                    "errors": formset_errors,
                    "form": self.formset,
                    "layout": self.layout,
                },
            )
        return mark_safe("")

    def render(self):
        return format_html(self.render_management_form() + "{}{}", self.render_errors(), self.render_forms())


class FormRenderer(BaseRenderer):
    """Default form renderer."""

    def __init__(self, form, *args, **kwargs):
        if not isinstance(form, BaseForm):
            raise BootstrapError('Parameter "form" should contain a valid Django Form.')
        self.form = form
        super().__init__(*args, **kwargs)

    def get_context_data(self):
        context = super().get_context_data()
        context["form"] = self.form
        return context

    def render_fields(self):
        rendered_fields = mark_safe("")
        kwargs = self.get_kwargs()
        for field in self.form:
            rendered_fields += render_field(field, **kwargs)
        return rendered_fields

    def get_fields_errors(self):
        form_errors = []
        for field in self.form:
            if not field.is_hidden and field.errors:
                form_errors += field.errors
        return form_errors

    def render_errors(self, type="all"):
        form_errors = None
        if type == "all":
            form_errors = self.get_fields_errors() + self.form.non_field_errors()
        elif type == "fields":
            form_errors = self.get_fields_errors()
        elif type == "non_fields":
            form_errors = self.form.non_field_errors()

        if form_errors:
            return render_template_file(
                "django_bootstrap5/form_errors.html",
                context={"errors": form_errors, "form": self.form, "layout": self.layout, "type": type},
            )

        return mark_safe("")

    def render(self):
        errors = self.render_errors(self.alert_error_type)
        fields = self.render_fields()
        return errors + fields


class FieldRenderer(BaseRenderer):
    """Default field renderer."""

    def __init__(self, field, *args, **kwargs):
        if not isinstance(field, BoundField):
            raise BootstrapError('Parameter "field" should contain a valid Django BoundField.')
        self.field = field
        super().__init__(*args, **kwargs)

        self.widget = field.field.widget
        self.is_multi_widget = isinstance(field.field.widget, MultiWidget)
        self.initial_attrs = self.widget.attrs.copy()
        self.help_text = text_value(field.help_text) if self.show_help and field.help_text else ""
        self.field_errors = [conditional_escape(text_value(error)) for error in field.errors]

        self.placeholder = text_value(kwargs.get("placeholder", self.default_placeholder))

        self.addon_before = kwargs.get("addon_before", self.widget.attrs.pop("addon_before", ""))
        self.addon_after = kwargs.get("addon_after", self.widget.attrs.pop("addon_after", ""))
        self.addon_before_class = kwargs.get(
            "addon_before_class", self.widget.attrs.pop("addon_before_class", "input-group-text")
        )
        self.addon_after_class = kwargs.get(
            "addon_after_class", self.widget.attrs.pop("addon_after_class", "input-group-text")
        )

        # These are set in Django or in the global BOOTSTRAP5 settings, and can be overwritten in the template
        error_css_class = kwargs.get("error_css_class", None)
        self.error_css_class = (
            getattr(field.form, "error_css_class", get_bootstrap_setting("error_css_class"))
            if error_css_class is None
            else error_css_class
        )

        required_css_class = kwargs.get("required_css_class", None)
        self.required_css_class = (
            getattr(field.form, "required_css_class", get_bootstrap_setting("required_css_class"))
            if required_css_class is None
            else required_css_class
        )
        if self.field.form.empty_permitted:
            self.required_css_class = ""

        bound_css_class = kwargs.get("bound_css_class", None)
        self.success_css_class = (
            getattr(field.form, "bound_css_class", get_bootstrap_setting("success_css_class"))
            if bound_css_class is None
            else bound_css_class
        )

    @property
    def is_floating(self):
        return (
            super().is_floating
            and self.can_widget_float(self.widget)
            and not self.addon_before
            and not self.addon_after
        )

    @property
    def default_placeholder(self):
        """Return default placeholder for field."""
        return self.field.label if get_bootstrap_setting("set_placeholder") else ""

    def restore_widget_attrs(self):
        self.widget.attrs = self.initial_attrs.copy()

    def get_widget_input_type(self, widget):
        """Return input type of widget, or None."""
        return widget.input_type if isinstance(widget, Input) else None

    def is_form_control_widget(self, widget=None):
        widget = widget or self.widget
        if isinstance(widget, Input):
            return self.get_widget_input_type(widget) in [
                "text",
                "number",
                "email",
                "url",
                "tel",
                "date",
                "time",
                "password",
            ]

        return isinstance(widget, Textarea)

    def can_widget_float(self, widget):
        """Return whether given widget can be set to `form-floating` behavior."""
        if self.is_form_control_widget(widget):
            return True
        if isinstance(widget, Select):
            return self.size == DEFAULT_SIZE and not isinstance(widget, (SelectMultiple, RadioSelect))
        return False

    def add_widget_class_attrs(self, widget=None):
        """Add class attribute to widget."""
        if widget is None:
            widget = self.widget
        size_prefix = None

        before = []
        classes = [widget.attrs.get("class", "")]

        if ReadOnlyPasswordHashWidget is not None and isinstance(widget, ReadOnlyPasswordHashWidget):
            before.append("form-control-static")
        elif isinstance(widget, Select):
            before.append("form-select")
            size_prefix = "form-select"
        elif isinstance(widget, CheckboxInput):
            before.append("form-check-input")
        elif isinstance(widget, (Input, Textarea)):
            input_type = self.get_widget_input_type(widget)
            if input_type == "range":
                before.append("form-range")
            else:
                before.append("form-control")
                if input_type == "color":
                    before.append("form-control-color")
                size_prefix = "form-control"

        if size_prefix:
            classes.append(get_size_class(self.size, prefix=size_prefix, skip=["xs", "md"]))

        if self.field.errors:
            if self.error_css_class:
                classes.append(self.error_css_class)
        elif self.field.form.is_bound:
            classes.append(self.success_css_class)

        classes = before + classes
        widget.attrs["class"] = merge_css_classes(*classes)

    def add_placeholder_attrs(self, widget=None):
        """Add placeholder attribute to widget."""
        if widget is None:
            widget = self.widget
        placeholder = widget.attrs.get("placeholder", self.placeholder)
        if placeholder and self.set_placeholder and is_widget_with_placeholder(widget):
            widget.attrs["placeholder"] = conditional_escape(strip_tags(placeholder))

    def add_widget_attrs(self):
        """Return HTML attributes for widget as dict."""
        if self.is_multi_widget:
            widgets = self.widget.widgets
        else:
            widgets = [self.widget]
        for widget in widgets:
            self.add_widget_class_attrs(widget)
            self.add_placeholder_attrs(widget)
            if isinstance(widget, (RadioSelect, CheckboxSelectMultiple)):
                widget.template_name = "django_bootstrap5/widgets/radio_select.html"
            elif isinstance(widget, ClearableFileInput):
                widget.template_name = "django_bootstrap5/widgets/clearable_file_input.html"

    def get_label_class(self, horizontal=False):
        """Return CSS class for label."""
        label_classes = [text_value(self.label_class)]
        if not self.show_label:
            label_classes.append("visually-hidden")
        else:
            if isinstance(self.widget, CheckboxInput):
                widget_label_class = "form-check-label"
            elif self.is_inline:
                widget_label_class = "visually-hidden"
            elif horizontal:
                widget_label_class = merge_css_classes(self.horizontal_label_class, "col-form-label")
            else:
                widget_label_class = "form-label"
            label_classes = [widget_label_class] + label_classes
        return merge_css_classes(*label_classes)

    def get_field_html(self):
        """Return HTML for field."""
        self.add_widget_attrs()
        field_html = self.field.as_widget(attrs=self.widget.attrs)
        self.restore_widget_attrs()
        return field_html

    def get_label_html(self, horizontal=False):
        """Return value for label."""
        label_html = "" if self.show_label == "skip" else self.field.label
        if label_html:
            label_html = render_label(
                label_html,
                label_for=self.field.id_for_label,
                label_class=self.get_label_class(horizontal=horizontal),
            )
        return label_html

    def get_help_html(self):
        """Return HTML for help text."""
        help_text = self.help_text or ""
        if help_text:
            return render_template_file(
                "django_bootstrap5/field_help_text.html",
                context={
                    "field": self.field,
                    "help_text": help_text,
                    "layout": self.layout,
                    "show_help": self.show_help,
                },
            )
        return ""

    def get_errors_html(self):
        """Return HTML for field errors."""
        field_errors = self.field_errors
        if field_errors:
            return render_template_file(
                "django_bootstrap5/field_errors.html",
                context={
                    "field": self.field,
                    "field_errors": field_errors,
                    "layout": self.layout,
                    "show_help": self.show_help,
                },
            )
        return ""

    def get_inline_field_class(self):
        """Return CSS class for inline field."""
        return self.inline_field_class or "col-12"

    def get_checkbox_classes(self):
        """Return CSS classes for checkbox."""
        classes = ["form-check"]
        if self.checkbox_style == "switch":
            classes.append("form-switch")
        if self.checkbox_layout == "inline":
            classes.append("form-check-inline")
        return merge_css_classes(*classes)

    def get_wrapper_classes(self):
        """Return classes for wrapper."""
        wrapper_classes = [self.wrapper_class]

        if self.is_floating:
            wrapper_classes.append("form-floating")

        if self.field.errors:
            wrapper_classes.append(self.error_css_class)
        elif self.field.form.is_bound:
            wrapper_classes.append(self.success_css_class)

        if self.field.field.required:
            wrapper_classes.append(self.required_css_class)

        if self.is_inline:
            wrapper_classes.append(self.get_inline_field_class())
        else:
            if self.is_horizontal:
                wrapper_classes.append("row")
            wrapper_classes.append("mb-3")

        return merge_css_classes(*wrapper_classes)

    def field_before_label(self):
        """Return whether field should be placed before label."""
        return isinstance(self.widget, CheckboxInput) or self.is_floating

    def render(self):
        if self.field.name in self.exclude.replace(" ", "").split(","):
            return ""
        if self.field.is_hidden:
            return text_value(self.field)

        field = self.get_field_html()
        if self.field_before_label():
            label = self.get_label_html()
            field = field + label
            label = mark_safe("")
            horizontal_class = merge_css_classes(self.horizontal_field_class, self.horizontal_field_offset_class)
        else:
            label = self.get_label_html(horizontal=self.is_horizontal)
            horizontal_class = self.horizontal_field_class

        if isinstance(self.widget, CheckboxInput):
            field = format_html(
                '<div class="{form_check_class}">{field}</div>',
                form_check_class=self.get_checkbox_classes(),
                field=field,
            )

        if self.is_form_control_widget():
            addon_before = (
                format_html('<span class="input-group-text">{}</span>', self.addon_before) if self.addon_before else ""
            )
            addon_after = (
                format_html('<span class="input-group-text">{}</span>', self.addon_after) if self.addon_after else ""
            )
            if addon_before or addon_after:
                field = format_html('<div class="input-group mb-3">{}{}{}</div>', addon_before, field, addon_after)

        field_with_help_and_errors = format_html("{}{}{}", field, self.get_help_html(), self.get_errors_html())
        if self.is_horizontal:
            field_with_help_and_errors = format_html(
                '<div class="{}">{}</div>', horizontal_class, field_with_help_and_errors
            )

        return format_html(
            '<{tag} class="{wrapper_classes}">{label}{field_with_help_and_errors}</{tag}>',
            tag=WRAPPER_TAG,
            wrapper_classes=self.get_wrapper_classes(),
            label=label,
            field_with_help_and_errors=field_with_help_and_errors,
        )



from django.test import TestCase

from django_bootstrap5.text import text_concat, text_value


class TextTestCase(TestCase):
    def test_text_value(self):
        self.assertEqual(text_value(""), "")
        self.assertEqual(text_value(" "), " ")
        self.assertEqual(text_value(None), "")
        self.assertEqual(text_value(1), "1")

    def test_text_concat(self):
        self.assertEqual(text_concat(1, 2), "12")
        self.assertEqual(text_concat(1, 2, separator="="), "1=2")
        self.assertEqual(text_concat(None, 2, separator="="), "2")



from django.contrib.messages import constants as DEFAULT_MESSAGE_LEVELS

from tests.base import BootstrapTestCase


class FakeMessage(object):
    """Follows the `django.contrib.messages.storage.base.Message` API."""

    level = None
    message = None
    extra_tags = None

    def __init__(self, level, message, extra_tags=None):
        self.level = level
        self.extra_tags = extra_tags
        self.message = message

    def __str__(self):
        return self.message


class MessagesTestCase(BootstrapTestCase):
    def test_bootstrap_messages(self):
        messages = [FakeMessage(DEFAULT_MESSAGE_LEVELS.WARNING, "hello")]
        html = self.render("{% bootstrap_messages messages %}", {"messages": messages})
        expected = (
            '<div class="alert alert-warning alert-dismissible fade show" role="alert">'
            "hello"
            '<button type="button" class="btn-close" data-bs-dismiss="alert" aria-label="Close"></button>'
            "</div>"
        )
        self.assertHTMLEqual(html, expected)

        messages = [FakeMessage(DEFAULT_MESSAGE_LEVELS.ERROR, "hello")]
        html = self.render("{% bootstrap_messages messages %}", {"messages": messages})
        expected = (
            '<div class="alert alert-danger alert-dismissible fade show" role="alert">'
            "hello"
            '<button type="button" class="btn-close" data-bs-dismiss="alert" aria-label="Close"></button>'
            "</div>"
        )
        self.assertHTMLEqual(html, expected)

        messages = [FakeMessage(None, "hello")]
        html = self.render("{% bootstrap_messages messages %}", {"messages": messages})
        expected = (
            '<div class="alert alert-info alert-dismissible fade show" role="alert">'
            "hello"
            '<button type="button" class="btn-close" data-bs-dismiss="alert" aria-label="Close"></button>'
            "</div>"
        )
        self.assertHTMLEqual(html, expected)

        messages = [FakeMessage(DEFAULT_MESSAGE_LEVELS.ERROR, "hello http://example.com")]
        html = self.render("{% bootstrap_messages messages %}", {"messages": messages})
        expected = (
            '<div class="alert alert-danger alert-dismissible fade show" role="alert">'
            "hello http://example.com"
            '<button type="button" class="btn-close" data-bs-dismiss="alert" aria-label="Close"></button>'
            "</div>"
        )
        self.assertHTMLEqual(html, expected)

        messages = [FakeMessage(DEFAULT_MESSAGE_LEVELS.ERROR, "hello\nthere")]
        html = self.render("{% bootstrap_messages messages %}", {"messages": messages})
        expected = (
            '<div class="alert alert-danger alert-dismissible fade show" role="alert">'
            "hello there"
            '<button type="button" class="btn-close" data-bs-dismiss="alert" aria-label="Close"></button>'
            "</div>"
        )
        self.assertHTMLEqual(html, expected)



from django import get_version
from django.template import engines
from django.test import TestCase

DJANGO3 = get_version() >= "3"


def html_39x27(html):
    """
    Return HTML string with &#39; (Django < 3) instead of &#x27; (Django >= 3).

    See https://docs.djangoproject.com/en/dev/releases/3.0/#miscellaneous
    """
    if not DJANGO3:
        return html.replace("&#x27;", "&#39;")
    return html


class BootstrapTestCase(TestCase):
    """TestCase with render function for template code."""

    def render(self, text, context=None, load_bootstrap=True):
        """Return rendered result of template with given context."""
        prefix = "{% load django_bootstrap5 %}" if load_bootstrap else ""
        template = engines["django"].from_string(f"{prefix}{text}")
        return template.render(context or {})



from importlib import import_module

from django.conf import settings

BOOTSTRAP5 = {"foo": "bar"}
BOOTSTRAP5_DEFAULTS = {
    "css_url": {
        "url": "https://cdn.jsdelivr.net/npm/bootstrap@5.0.0-beta3/dist/css/bootstrap.min.css",
        "integrity": "sha384-eOJMYsd53ii+scO/bJGFsiCZc+5NDVN2yr8+0RDqr0Ql0h+rP48ckxlpbzKgwra6",
        "crossorigin": "anonymous",
    },
    "javascript_url": {
        "url": "https://cdn.jsdelivr.net/npm/bootstrap@5.0.0-beta3/dist/js/bootstrap.bundle.min.js",
        "integrity": "sha384-JEW9xMcG8R+pH31jmWH6WWP0WintQrMb4s7ZOdauHnUtxwoG2vI5DkLtS3qm9Ekf",
        "crossorigin": "anonymous",
    },
    "theme_url": None,
    "javascript_in_head": False,
    "horizontal_label_class": "col-sm-2",
    "horizontal_field_class": "col-sm-10",
    "horizontal_field_offset_class": "offset-sm-2",
    "set_placeholder": True,
    "checkbox_layout": None,
    "checkbox_style": None,
    "required_css_class": "",
    "error_css_class": "is-invalid",
    "success_css_class": "is-valid",
    "formset_renderers": {"default": "django_bootstrap5.renderers.FormsetRenderer"},
    "form_renderers": {"default": "django_bootstrap5.renderers.FormRenderer"},
    "field_renderers": {
        "default": "django_bootstrap5.renderers.FieldRenderer",
    },
}


def get_bootstrap_setting(name, default=None):
    """Read a setting."""
    # Start with a copy of default settings
    BOOTSTRAP5 = BOOTSTRAP5_DEFAULTS.copy()

    # Override with user settings from settings.py
    BOOTSTRAP5.update(getattr(settings, "BOOTSTRAP5", {}))

    return BOOTSTRAP5.get(name, default)


def javascript_url():
    """Return the full url to the Bootstrap JavaScript file."""
    return get_bootstrap_setting("javascript_url")


def css_url():
    """Return the full url to the Bootstrap CSS file."""
    return get_bootstrap_setting("css_url")


def theme_url():
    """Return the full url to the theme CSS file."""
    return get_bootstrap_setting("theme_url")


def get_renderer(renderers, **kwargs):
    layout = kwargs.get("layout", "")
    path = renderers.get(layout, renderers["default"])
    mod, cls = path.rsplit(".", 1)
    return getattr(import_module(mod), cls)


def get_formset_renderer(**kwargs):
    renderers = get_bootstrap_setting("formset_renderers")
    return get_renderer(renderers, **kwargs)


def get_form_renderer(**kwargs):
    renderers = get_bootstrap_setting("form_renderers")
    return get_renderer(renderers, **kwargs)


def get_field_renderer(**kwargs):
    renderers = get_bootstrap_setting("field_renderers")
    return get_renderer(renderers, **kwargs)



from tests.base import BootstrapTestCase


class BootstrapLabelTestCase(BootstrapTestCase):
    def test_bootstrap_label(self):
        self.assertHTMLEqual(
            self.render('{% bootstrap_label "Subject" %}'),
            "<label>Subject</label>",
        )
        self.assertHTMLEqual(
            self.render("{% bootstrap_label \"Subject\" label_for='subject' %}"),
            '<label for="subject">Subject</label>',
        )
        self.assertHTMLEqual(
            self.render("{% bootstrap_label \"Subject\" label_class='label-class' %}"),
            '<label class="label-class">Subject</label>',
        )
        self.assertHTMLEqual(
            self.render("{% bootstrap_label \"Subject\" label_title='label-title' %}"),
            '<label title="label-title">Subject</label>',
        )



from django import forms

from tests.base import BootstrapTestCase


class CheckboxTestForm(forms.Form):
    test = forms.BooleanField()


class InputTypeCheckboxTestCase(BootstrapTestCase):
    def test_input_type_checkbox(self):
        """Test field with checkbox widget."""
        self.assertHTMLEqual(
            self.render("{% bootstrap_field form.test %}", context={"form": CheckboxTestForm()}),
            (
                '<div class="django_bootstrap5-req mb-3">'
                '<div class="form-check">'
                '<input class="form-check-input" id="id_test" name="test" required type="checkbox">'
                '<label class="form-check-label" for="id_test">'
                "Test"
                "</label>"
            ),
        )

    def test_input_type_checkbox_style_switch(self):
        """Test field with checkbox widget, style switch."""
        self.assertHTMLEqual(
            self.render(
                '{% bootstrap_field form.test checkbox_style="switch" %}', context={"form": CheckboxTestForm()}
            ),
            (
                '<div class="django_bootstrap5-req mb-3">'
                '<div class="form-check form-switch">'
                '<input class="form-check-input" id="id_test" name="test" required type="checkbox">'
                '<label class="form-check-label" for="id_test">'
                "Test"
                "</label>"
            ),
        )

    def test_bootstrap_field_checkbox_horizontal(self):
        """Test field with checkbox widget, layout horizontal."""
        self.assertHTMLEqual(
            self.render("{% bootstrap_field form.test layout='horizontal' %}", context={"form": CheckboxTestForm()}),
            (
                '<div class="django_bootstrap5-req mb-3 row">'
                '<div class="col-sm-10 offset-sm-2">'
                '<div class="form-check">'
                '<input class="form-check-input" id="id_test" name="test" required type="checkbox">'
                '<label class="form-check-label" for="id_test">'
                "Test"
                "</label>"
                "</div>"
            ),
        )



from django_bootstrap5.core import get_bootstrap_setting
from tests.base import BootstrapTestCase


class MediaTestCase(BootstrapTestCase):
    expected_bootstrap_css = (
        '<link href="{url}" integrity="{integrity}" crossorigin="{crossorigin}" rel="stylesheet">'.format(
            **get_bootstrap_setting("css_url")
        )
    )
    expected_bootstrap_js = '<script src="{url}" integrity="{integrity}" crossorigin="{crossorigin}"></script>'.format(
        **get_bootstrap_setting("javascript_url")
    )

    def test_bootstrap_javascript_tag(self):
        self.assertHTMLEqual(
            self.render("{% bootstrap_javascript %}"),
            self.expected_bootstrap_js,
        )

    def test_bootstrap_css_tag(self):
        self.assertHTMLEqual(
            self.render("{% bootstrap_css %}"),
            self.expected_bootstrap_css,
        )

    def test_bootstrap_css_tag_with_theme(self):
        with self.settings(BOOTSTRAP5={"theme_url": "//example.com/theme.css"}):
            self.assertHTMLEqual(
                self.render("{% bootstrap_css %}"),
                self.expected_bootstrap_css + '<link rel="stylesheet" href="//example.com/theme.css">',
            )

    def test_bootstrap_setting_filter(self):
        self.assertEqual(self.render('{{ "required_css_class"|bootstrap_setting }}'), "django_bootstrap5-req")
        self.assertEqual(
            self.render('{% if "javascript_in_head"|bootstrap_setting %}head{% else %}body{% endif %}'), "head"
        )



from urllib.parse import parse_qs, urlparse, urlunparse

from django.template.loader import get_template
from django.utils.encoding import force_str
from django.utils.http import urlencode
from django.utils.safestring import mark_safe


def render_template_file(template, context=None):
    """Return rendered template file, with given context as input."""
    template = get_template(template)
    return template.render(context)


def url_replace_param(url, name, value):
    """Replace a GET parameter in a URL."""
    url_components = urlparse(force_str(url))

    params = parse_qs(url_components.query)

    if value is None:
        del params[name]
    else:
        params[name] = value

    return mark_safe(
        urlunparse(
            [
                url_components.scheme,
                url_components.netloc,
                url_components.path,
                url_components.params,
                urlencode(params, doseq=True),
                url_components.fragment,
            ]
        )
    )


def get_url_attrs(url, attr_name):
    """
    Return dictionary with attributes for HTML tag, updated with key `attr_name` with value `url`.

    Parameter `url` is either a string or a dict of attrs with the key `url`.
    Parameter `attr_key` is the name for the url value in the results.
    """
    url_attrs = {}
    if isinstance(url, str):
        url = {"url": url}
    url_attrs.update(url)
    url_attrs[attr_name] = url_attrs.pop("url")
    return url_attrs



from django.forms import EmailInput, NumberInput, PasswordInput, RadioSelect, Textarea, TextInput, URLInput

try:
    # If Django is set up without a database, importing this widget gives RuntimeError
    from django.contrib.auth.forms import ReadOnlyPasswordHashWidget
except RuntimeError:
    ReadOnlyPasswordHashWidget = None


class RadioSelectButtonGroup(RadioSelect):
    """A RadioSelect that renders as a horizontal button groep."""

    template_name = "django_bootstrap5/widgets/radio_select_button_group.html"


def is_widget_with_placeholder(widget):
    """Return whether this widget can have a placeholder."""
    return isinstance(widget, (TextInput, Textarea, NumberInput, EmailInput, URLInput, PasswordInput))



from django.template.defaultfilters import capfirst
from django.utils.html import format_html
from django.utils.translation import gettext as _

from .css import merge_css_classes
from .html import render_tag

ALERT_TYPES = ["primary", "secondary", "success", "danger", "warning", "info", "light", "dark"]


def render_alert(content, alert_type="info", dismissible=True, extra_classes=""):
    """Render a Bootstrap alert."""
    button = ""
    if alert_type not in ALERT_TYPES:
        raise ValueError(f"Value {alert_type} is not a valid alert type. Please choose from {', '.join(ALERT_TYPES)}.")
    css_classes = [f"alert alert-{alert_type}"]
    if dismissible:
        css_classes.append("alert-dismissible fade show")
        close = capfirst(_("close"))
        button = f'<button type="button" class="btn-close" data-bs-dismiss="alert" aria-label="{close}"></button>'
    css_classes = merge_css_classes(*css_classes)
    return render_tag(
        "div",
        attrs={"class": css_classes, "role": "alert"},
        content=format_html("{content}" + button, content=content),
    )



from django.test import TestCase

from django_bootstrap5.html import render_tag


class HtmlTestCase(TestCase):
    def test_render_tag(self):
        self.assertEqual(render_tag("span"), "<span></span>")
        self.assertEqual(render_tag("span", content="foo"), "<span>foo</span>")
        self.assertEqual(render_tag("span", attrs={"bar": 123}, content="foo"), '<span bar="123">foo</span>')



import tempfile

from django import forms
from django.core.files.uploadedfile import SimpleUploadedFile
from django.test import override_settings

from tests.base import BootstrapTestCase
from tests.image import get_test_image


class FileFieldTestForm(forms.Form):
    test = forms.FileField()


class ClearableFileInputTestForm(forms.Form):
    test = forms.FileField(widget=forms.ClearableFileInput, required=False)


class ImageFieldTestForm(forms.Form):
    test = forms.ImageField()


class InputTypeFileTestCase(BootstrapTestCase):
    def test_input_type_file(self):
        self.assertHTMLEqual(
            self.render("{% bootstrap_field form.test %}", context={"form": FileFieldTestForm()}),
            (
                '<div class="django_bootstrap5-req mb-3">'
                '<label for="id_test" class="form-label">Test</label>'
                '<input class="form-control" id="id_test" name="test" required type="file">'
                "</div>"
            ),
        )

    def test_clearable_file_input(self):
        self.assertHTMLEqual(
            self.render("{% bootstrap_field form.test %}", context={"form": ClearableFileInputTestForm()}),
            (
                '<div class="mb-3">'
                '<label for="id_test" class="form-label">Test</label>'
                '<input class="form-control" id="id_test" name="test" type="file">'
                "</div>"
            ),
        )

    def test_clearable_file_input_post(self):
        self.assertHTMLEqual(
            self.render(
                "{% bootstrap_field form.test %}",
                context={"form": ClearableFileInputTestForm({}, {"test": SimpleUploadedFile("test.txt", b"test")})},
            ),
            (
                '<div class="django_bootstrap5-bound mb-3">'
                '<label class="form-label" for="id_test">Test</label>'
                '<input type="file" name="test" class="form-control django_bootstrap5-bound" id="id_test">'
                "</div>"
            ),
        )

    def test_image_field(self):
        self.assertHTMLEqual(
            self.render("{% bootstrap_field form.test %}", context={"form": ImageFieldTestForm()}),
            (
                '<div class="django_bootstrap5-req mb-3">'
                '<label for="id_test" class="form-label">Test</label>'
                '<input accept="image/*" class="form-control" id="id_test" name="test" required type="file">'
                "</div>"
            ),
        )

    @override_settings(MEDIA_ROOT=tempfile.gettempdir())
    def test_image_field_post(self):
        form = ImageFieldTestForm({}, {"test": get_test_image()})
        self.assertHTMLEqual(
            self.render(
                "{% bootstrap_field form.test %}",
                context={"form": form},
            ),
            (
                '<div class="django_bootstrap5-bound django_bootstrap5-req mb-3">'
                '<label class="form-label" for="id_test">Test</label>'
                '<input type="file" name="test" accept="image/*" class="form-control django_bootstrap5-bound"'
                ' required id="id_test">'
                "</div>"
            ),
        )



from django import forms

from django_bootstrap5.exceptions import BootstrapError
from tests.base import BootstrapTestCase


class TestForm(forms.Form):
    subject = forms.CharField(required=True)
    date = forms.DateField()


TestFormSet = forms.formset_factory(TestForm)


class BootstrapFormSetTestCase(BootstrapTestCase):
    def test_formset(self):
        formset = TestFormSet()
        html = self.render("{% bootstrap_formset formset %}", {"formset": formset})
        self.assertHTMLEqual(
            html,
            """
<input type="hidden" name="form-TOTAL_FORMS" value="1" id="id_form-TOTAL_FORMS">
<input type="hidden" name="form-INITIAL_FORMS" value="0" id="id_form-INITIAL_FORMS">
<input type="hidden" name="form-MIN_NUM_FORMS" value="0" id="id_form-MIN_NUM_FORMS">
<input type="hidden" name="form-MAX_NUM_FORMS" value="1000" id="id_form-MAX_NUM_FORMS">
<div class="mb-3">
<label class="form-label" for="id_form-0-subject">Subject</label>
<input type="text" name="form-0-subject" class="form-control" placeholder="Subject" id="id_form-0-subject">
</div>
<div class="mb-3">
<label class="form-label" for="id_form-0-date">Date</label>
<input type="text" name="form-0-date" class="form-control" placeholder="Date" id="id_form-0-date">
</div>
            """,
        )

    def test_formset_post(self):
        formset = TestFormSet({})
        html = self.render("{% bootstrap_formset_errors formset %}", {"formset": formset})
        self.assertHTMLEqual(
            html,
            """
<ul class="list-unstyled text-danger">
<li>
    ManagementForm data is missing or has been tampered with.
    Missing fields: form-TOTAL_FORMS, form-INITIAL_FORMS.
    You may need to file a bug report if the issue persists.
</li>
</ul>
            """,
        )

    def test_illegal_formset(self):
        with self.assertRaises(BootstrapError):
            self.render("{% bootstrap_formset formset %}", {"formset": "illegal"})



from django.test import TestCase
from django.utils.safestring import mark_safe

from django_bootstrap5.components import render_alert
from django_bootstrap5.exceptions import BootstrapError
from django_bootstrap5.forms import render_button


class AlertsTestCase(TestCase):
    def test_render_alert_without_type(self):
        self.assertEqual(
            render_alert("content"),
            (
                '<div class="alert alert-info alert-dismissible fade show" role="alert">'
                "content"
                '<button type="button" class="btn-close" data-bs-dismiss="alert" aria-label="Close"></button>'
                "</div>"
            ),
        )

    def test_render_alert_with_type(self):
        self.assertEqual(
            render_alert("content", alert_type="danger"),
            (
                '<div class="alert alert-danger alert-dismissible fade show" role="alert">'
                "content"
                '<button type="button" class="btn-close" data-bs-dismiss="alert" aria-label="Close"></button>'
                "</div>"
            ),
        )

    def test_render_alert_with_safe_content(self):
        self.assertEqual(
            render_alert(mark_safe('This is <a href="https://example.com" class="alert-link">a safe link</a>!')),
            (
                '<div class="alert alert-info alert-dismissible fade show" role="alert">'
                'This is <a href="https://example.com" class="alert-link">a safe link</a>!'
                '<button type="button" class="btn-close" data-bs-dismiss="alert" aria-label="Close"></button>'
                "</div>"
            ),
        )

    def test_render_alert_with_unsafe_content(self):
        self.assertEqual(
            render_alert("This is <b>unsafe</b>!"),
            (
                '<div class="alert alert-info alert-dismissible fade show" role="alert">'
                "This is &lt;b&gt;unsafe&lt;/b&gt;!"
                '<button type="button" class="btn-close" data-bs-dismiss="alert" aria-label="Close"></button>'
                "</div>"
            ),
        )


class ButtonsTestCase(TestCase):
    def test_button(self):
        self.assertEqual(render_button("button"), '<button class="btn btn-primary">button</button>')

    def test_button_with_illegal_type(self):
        try:
            self.assertEqual(
                render_button("button", button_type="illegal"), '<button class="btn btn-primary">button</button>'
            )
        except BootstrapError as e:
            self.assertEqual(
                str(e),
                'Parameter "button_type" should be "submit", "reset", "button", "link" or empty ("illegal" given).',
            )



from django import forms

from django_bootstrap5.exceptions import BootstrapError

from .base import BootstrapTestCase, html_39x27


class XssTestForm(forms.Form):
    xss_field = forms.CharField(label='XSS" onmouseover="alert(\'Hello, XSS\')" foo="', max_length=100)


class SubjectTestForm(forms.Form):
    subject = forms.CharField(
        max_length=100,
        help_text="my_help_text",
        required=True,
        widget=forms.TextInput(attrs={"placeholder": "placeholdertest"}),
    )


class FieldTestCase(BootstrapTestCase):
    def test_illegal_field(self):
        with self.assertRaises(BootstrapError):
            self.render("{% bootstrap_field field %}", {"field": "illegal"})

    def test_show_help(self):
        html = self.render("{% bootstrap_field form.subject %}", {"form": SubjectTestForm()})
        self.assertIn("my_help_text", html)
        self.assertNotIn("<i>my_help_text</i>", html)
        html = self.render("{% bootstrap_field form.subject show_help=False %}", {"form": SubjectTestForm()})
        self.assertNotIn("my_help_text", html)

    def test_placeholder(self):
        html = self.render("{% bootstrap_field form.subject %}", {"form": SubjectTestForm()})
        self.assertIn('type="text"', html)
        self.assertIn('placeholder="placeholdertest"', html)

    def test_xss_field(self):
        html = self.render("{% bootstrap_field form.xss_field %}", {"form": XssTestForm()})
        self.assertIn('type="text"', html)
        self.assertIn(html_39x27(">XSS&quot; onmouseover=&quot;alert(&#x27;Hello, XSS&#x27;)&quot; foo=&quot;<"), html)
        self.assertIn(
            html_39x27('placeholder="XSS&quot; onmouseover=&quot;alert(&#x27;Hello, XSS&#x27;)&quot; foo=&quot;"'), html
        )

    def test_empty_permitted(self):
        """If a form has empty_permitted, no fields should get the CSS class for required."""
        form = SubjectTestForm()

        html = self.render("{% bootstrap_field form.subject %}", {"form": form})
        self.assertIn("django_bootstrap5-req", html)

        form.empty_permitted = True
        html = self.render("{% bootstrap_field form.subject %}", {"form": form})
        self.assertNotIn("django_bootstrap5-req", html)

    def test_size(self):
        def _test_size(param, klass):
            html = self.render('{% bootstrap_field form.subject size="' + param + '" %}', {"form": SubjectTestForm()})
            self.assertIn(klass, html)

        def _test_size_medium(param):
            html = self.render('{% bootstrap_field form.subject size="' + param + '" %}', {"form": SubjectTestForm()})
            self.assertNotIn("form-control-lg", html)
            self.assertNotIn("form-control-sm", html)
            self.assertNotIn("form-control-md", html)

        _test_size("sm", "form-control-sm")
        _test_size("lg", "form-control-lg")
        _test_size_medium("md")
        _test_size_medium("")

    def test_label(self):
        self.assertEqual(
            self.render('{% bootstrap_label "foobar" label_for="subject" %}'),
            '<label for="subject">foobar</label>',
        )



from django.test import TestCase


class VersionTestCase(TestCase):
    """Test presence of package version."""

    def test_version(self):
        import django_bootstrap5

        version = django_bootstrap5.__version__
        version_parts = version.split(".")
        self.assertTrue(len(version_parts) >= 3)



from django.forms.utils import flatatt
from django.utils.html import format_html

from django_bootstrap5.text import text_value
from django_bootstrap5.utils import get_url_attrs


def render_script_tag(url):
    """Build a script tag."""
    return render_tag("script", get_url_attrs(url, attr_name="src"))


def render_link_tag(url):
    """Build a link tag."""
    attrs = get_url_attrs(url, attr_name="href")
    attrs["rel"] = "stylesheet"
    return render_tag("link", attrs=attrs, close=False)


def render_tag(tag, attrs=None, content=None, close=True):
    """Render an HTML tag."""
    attrs_string = flatatt(attrs) if attrs else ""
    builder = "<{tag}" + attrs_string + ">{content}"
    content_string = text_value(content)
    if content_string or close:
        builder += "</{tag}>"
    return format_html(builder, tag=tag, content=content_string)



from django.core.paginator import Paginator

from django_bootstrap5.utils import url_replace_param
from tests.base import BootstrapTestCase


class PaginatorTestCase(BootstrapTestCase):
    def test_url_replace_param(self):
        self.assertEqual(url_replace_param("/foo/bar?baz=foo", "baz", "yohoo"), "/foo/bar?baz=yohoo")
        self.assertEqual(url_replace_param("/foo/bar?baz=foo", "baz", None), "/foo/bar")
        self.assertEqual(url_replace_param("/foo/bar#id", "baz", "foo"), "/foo/bar?baz=foo#id")

    def bootstrap_pagination(self, page, extra=""):
        """Helper to test bootstrap_pagination tag."""
        return self.render(f"{{% bootstrap_pagination page {extra} %}}", {"page": page})

    def test_paginator(self):
        objects = ["john", "paul", "george", "ringo"]
        paginator = Paginator(objects, 2)

        html = self.bootstrap_pagination(paginator.page(2), extra='url="/projects/?foo=bar"')
        self.assertHTMLEqual(
            html,
            """
<nav>
    <ul class="pagination">
        <li class="page-item"><a class="page-link" href="/projects/?foo=bar&page=1">&laquo;</a></li>
        <li class="page-item"><a class="page-link" href="/projects/?foo=bar&page=1">1</a></li>
        <li class="page-item active"><a class="page-link" href="#">2</a></li>
        <li class="page-item disabled"><a class="page-link" href="#">&raquo;</a></li>
    </ul>
</nav>
    """,
        )
        self.assertIn("/projects/?foo=bar&page=1", html)
        self.assertNotIn("/projects/?foo=bar&page=2", html)

        html = self.bootstrap_pagination(paginator.page(2), extra='url="/projects/#id"')
        self.assertIn("/projects/?page=1#id", html)
        self.assertNotIn("/projects/?page=2#id", html)

        html = self.bootstrap_pagination(paginator.page(2), extra='url="/projects/?page=3#id"')
        self.assertIn("/projects/?page=1#id", html)
        self.assertNotIn("/projects/?page=2#id", html)

        html = self.bootstrap_pagination(paginator.page(2), extra='url="/projects/?page=3" extra="id=20"')
        self.assertIn("/projects/?page=1&id=20", html)
        self.assertNotIn("/projects/?page=2&id=20", html)



from django import forms
from django.contrib.admin.widgets import AdminSplitDateTime

from tests.base import BootstrapTestCase


class DateTimeTestForm(forms.Form):
    test = forms.SplitDateTimeField(widget=AdminSplitDateTime())


class DateTimeTestCase(BootstrapTestCase):
    def test_input_type_admin_split_date_time(self):
        """Test field with AdminSplitDateTime widget."""
        self.assertHTMLEqual(
            self.render("{% bootstrap_field form.test %}", context={"form": DateTimeTestForm()}),
            (
                '<div class="django_bootstrap5-req mb-3">'
                '<label class="form-label" for="id_test_0">Test</label>'
                '<p class="datetime">'
                "Date: "
                '<input type="text" name="test_0" class="form-control vDateField" size="10"'
                ' placeholder="Test" required id="id_test_0">'
                "<br>"
                "Time: "
                '<input type="text" name="test_1" class="form-control vTimeField" size="8"'
                ' placeholder="Test" required id="id_test_1">'
                "</p>"
                "</div>"
            ),
        )

    def test_input_type_admin_split_date_time_horizontal(self):
        """Test field with AdminSplitDateTime widget with layout horizontal."""
        self.assertHTMLEqual(
            self.render('{% bootstrap_field form.test layout="horizontal" %}', context={"form": DateTimeTestForm()}),
            (
                '<div class="django_bootstrap5-req row mb-3">'
                '<label class="col-form-label col-sm-2" for="id_test_0">Test</label>'
                '<div class="col-sm-10">'
                '<p class="datetime">'
                "Date: "
                '<input type="text" name="test_0" class="form-control vDateField" size="10"'
                ' placeholder="Test" required id="id_test_0">'
                "<br>"
                "Time: "
                '<input type="text" name="test_1" class="form-control vTimeField" size="8"'
                ' placeholder="Test" required id="id_test_1">'
                "</p>"
                "</div>"
                "</div>"
            ),
        )

    def test_input_type_admin_split_date_time_floating(self):
        """Test field with AdminSplitDateTime widget with layout floating."""
        self.assertHTMLEqual(
            self.render('{% bootstrap_field form.test layout="floating" %}', context={"form": DateTimeTestForm()}),
            (
                '<div class="django_bootstrap5-req mb-3">'
                '<label class="form-label" for="id_test_0">Test</label>'
                '<p class="datetime">'
                "Date: "
                '<input type="text" name="test_0" class="form-control vDateField" size="10"'
                ' placeholder="Test" required id="id_test_0">'
                "<br>"
                "Time: "
                '<input type="text" name="test_1" class="form-control vTimeField" size="8"'
                ' placeholder="Test" required id="id_test_1">'
                "</p>"
                "</div>"
            ),
        )



from django import forms

from tests.base import BootstrapTestCase


class ColorTestForm(forms.Form):
    test = forms.CharField(widget=forms.TextInput(attrs={"type": "color"}))


class InputTypeColorTestCase(BootstrapTestCase):
    def test_input_type_color(self):
        """Test field with input widget with type `color`."""
        self.assertHTMLEqual(
            self.render("{% bootstrap_field form.test %}", context={"form": ColorTestForm()}),
            (
                '<div class="django_bootstrap5-req mb-3">'
                '<label for="id_test" class="form-label">Test</label>'
                '<input class="form-control form-control-color" '
                'id="id_test" name="test" placeholder="Test" required type="color">'
                "</div>"
            ),
        )

    def test_input_type_color_horizontal(self):
        """Test field with input widget with type `color` in horizontal layout."""
        self.assertHTMLEqual(
            self.render('{% bootstrap_field form.test layout="horizontal" %}', context={"form": ColorTestForm()}),
            (
                '<div class="django_bootstrap5-req row mb-3">'
                '<label for="id_test" class="col-form-label col-sm-2">Test</label>'
                '<div class="col-sm-10">'
                '<input class="form-control form-control-color" id="id_test"'
                ' name="test" placeholder="Test" required type="color">'
                "</div>"
                "</div>"
            ),
        )

    def test_input_type_color_floating(self):
        """Test field with input widget with type `color` in floating layout."""
        self.assertHTMLEqual(
            self.render('{% bootstrap_field form.test layout="floating" %}', context={"form": ColorTestForm()}),
            (
                '<div class="django_bootstrap5-req mb-3">'
                '<label for="id_test" class="form-label">Test</label>'
                '<input class="form-control form-control-color" id="id_test"'
                ' name="test" placeholder="Test" required type="color">'
                "</div>"
            ),
        )



from django_bootstrap5.css import merge_css_classes
from django_bootstrap5.exceptions import BootstrapError
from django_bootstrap5.text import text_value

SIZE_XS = "xs"
SIZE_SM = "sm"
SIZE_MD = "md"
SIZE_LG = "lg"
SIZES = [SIZE_SM, SIZE_MD, SIZE_LG]
DEFAULT_SIZE = SIZE_MD


def parse_size(value, default=None):
    """Return size if it is valid, default size if size is empty, or throws exception."""
    size = text_value(value or default)
    if size not in SIZES:
        valid_sizes = ", ".join(SIZES)
        raise BootstrapError(f'Invalid value "{size}" for parameter "size" (valid values are {valid_sizes}).')
    return size


def get_size_class(size, prefix, *, default=None, skip=None):
    """Return CSS class for size with given prefix, unless size needs to be skipped."""
    size = parse_size(size, default=default)
    if skip:
        if isinstance(skip, str):
            skip = merge_css_classes(skip).split(" ")
        if size in skip:
            return ""
    return f"{prefix}-{size}"
